---------------------------------------------------------------------------
-- - - - - - - - - - - - - - - - Room object - - - - - - - - - - - - - - -- 
---------------------------------------------------------------------------

-- Room is a Node-like object.

--  * Has unique id
--  * Keeps track of neighbouring rooms by keeping boolean value 
--  * at index representing neighbours' id's

rawset(_G, "Room", { id, neighbours, center, hasSpring , hasBoss })
Room.__index = Room

function Room:new(id, r, c, w, h)
	local room = {
		id=id,
		r=r, -- Row (Down=positive, Up=negative)
		c=c, -- Column (Right=positive, left=negative)
		w=w,
		h=h,
		neighbours = {},
		center = {},
		hasSpring = false,
		hasBoss = false
	}

	setmetatable(room, Room)

	return room
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Room:addNeighbour(other)
	table.insert(self.neighbours, other)
end
  
-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Room:hasNeighbours()
	return #self.neighbours>1
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Room:setCenter(r,c)
	self.center = {r, c}
end

function Room:getCorners()
	-- Return 4 corner points
	-- 1--2
	-- |  |
	-- |  |
	-- 3--4
	return {
			{self.r, self.c}, {self.r, self.h},
			{self.w, self.c}, {self.w, self.h}
			}
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

function Room:distanceTo(other)
	return R_PointToDist2(self.center[1]*FRACUNIT, self.center[2]*FRACUNIT, 
			other.center[1]*FRACUNIT, other.center[2]*FRACUNIT)/FRACUNIT
end