---------------------------------------------------------------------------
-- - - - - - - - - - - - - - - - Tile object - - - - - - - - - - - - - - -- 
---------------------------------------------------------------------------

-- Tile objects:

--  * Keeps track of room association, if not in room (default): roomId = 0
--  * Has graphic symbol to represent what kind of tile this is in a level

rawset(_G, "Tile", {class, roomId})
Tile.__index = Tile

Tile.WALL = 0 -- Impassible
Tile.WALK = 1 -- Walkable
Tile.CORR = 2 -- Corridor passageway
Tile.SPAWN = 3 -- Player Int Spawn
Tile.SPRING = 4 -- Spring (Ascending)
Tile.RING = 5 -- Ring collection
Tile.ITEM = 6 -- Item
Tile.KEY = 7 -- Key space
Tile.ENEMY = 8 -- Enemy
Tile.BOSS = 9 -- Boss
Tile.PIT = 10 -- Death Pit

function Tile:new(t)
	local tile = {}
	tile.class = t
	tile.roomId = 0
	tile.enemyId = 0

	setmetatable(tile, Tile)

	return tile
end

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### --

function Tile:isWall() 
	return (self.class == Tile.WALL)
end

function Tile:isWalkable()
	return not ((self.class == Tile.WALL) or (self.class == Tile.PIT))
end