-- LJ Sonic

rawset(_G, "serialLib", {})

serialLib.bytesToString = function(bytes)
	local chars = {}
	local j = 1

	for i = 1, #bytes / 1024
		chars[i] = string.char(unpack(bytes, j, j + 1023))
		j = j + 1024
	end

	chars[#bytes / 1024 + 1] = string.char(unpack(bytes, j, #bytes))

	return table.concat(chars)
end

serialLib.stringToBytes = function(s)
	local bytes = {}
	local j = 1

	for i = 1, #s / 1024
		local chunk = {s:byte(j, j + 1024 - 1)}
		for i = 1, 1024
			bytes[j] = chunk[i]
			j = j + 1
		end
	end

	local chunk = {s:byte(j, #s)}
	for i = 1, 1024
		bytes[j] = chunk[i]
		j = j + 1
	end

	return bytes
end

serialLib.createByteStream = function(bytes)
	return {
		bytes = bytes or {},
		byteoffset = 1,
		bitoffset = 0
	}
end

serialLib.readByte = function(stream)
	local bytes = stream.bytes

	local highbyteoffset = stream.byteoffset
	local highbitoffset = stream.bitoffset

	stream.byteoffset = highbyteoffset + 1

	if highbitoffset ~= 0
		local byte = (bytes[highbyteoffset]
			& (255 >> highbitoffset)) << highbitoffset

		local lowbitoffset = 8 - highbitoffset
		byte = byte | ((bytes[stream.byteoffset]
			& (255 << lowbitoffset)) >> lowbitoffset)

		return byte
	else
		return bytes[highbyteoffset]
	end
end
local readByte = serialLib.readByte

serialLib.writeByte = function(stream, byte)
	local bytes = stream.bytes

	local highbyteoffset = stream.byteoffset
	local highbitoffset = stream.bitoffset

	stream.byteoffset = highbyteoffset + 1

	if highbitoffset ~= 0
		bytes[highbyteoffset] = bytes[highbyteoffset]
			| ((byte & (255 << highbitoffset)) >> highbitoffset)

		local lowbitoffset = 8 - highbitoffset
		bytes[stream.byteoffset] = (byte
			& (255 >> lowbitoffset)) << lowbitoffset
	else
		bytes[highbyteoffset] = byte
	end
end
local writeByte = serialLib.writeByte

function serialLib.readUInt(stream, len)
	local bytes = stream.bytes
	local byteoffset = stream.byteoffset
	local bitoffset = stream.bitoffset

	if bitoffset + len < 8 and bitoffset ~= 0
		stream.bitoffset = bitoffset + len
		return (bytes[byteoffset] >> (8 - bitoffset - len)) & ((1 << len) - 1)
	end

	local n = 0

	if bitoffset ~= 0
		local space = 8 - bitoffset
		len = len - space
		n = (bytes[byteoffset] & ((1 << space) - 1)) << len

		byteoffset = byteoffset + 1
		bitoffset = 0
	end

	while len >= 8
		len = len - 8
		n = n | (bytes[byteoffset] << len)
		byteoffset = byteoffset + 1
	end

	if len ~= 0
		n = n | (bytes[byteoffset] >> (8 - len))
		bitoffset = len
	end

	stream.byteoffset = byteoffset
	stream.bitoffset = bitoffset

	return n
end
local readUInt = serialLib.readUInt

function serialLib.writeUInt(stream, len, n)
	local bytes = stream.bytes
	local byteoffset = stream.byteoffset
	local bitoffset = stream.bitoffset

	if bitoffset + len < 8 and bitoffset ~= 0
		bytes[byteoffset] = bytes[byteoffset] | (n << (8 - bitoffset - len))
		stream.bitoffset = bitoffset + len
		return
	end

	if bitoffset ~= 0
		local space = 8 - bitoffset
		len = len - space
		bytes[byteoffset] = bytes[byteoffset] | (n >> len)

		byteoffset = byteoffset + 1
		bitoffset = 0
	end

	while len >= 8
		len = len - 8
		bytes[byteoffset] = (n >> len) & 255
		byteoffset = byteoffset + 1
	end

	if len ~= 0
		bytes[byteoffset] = (n << (8 - len)) & 255
		bitoffset = len
	end

	stream.byteoffset = byteoffset
	stream.bitoffset = bitoffset
end
local writeUInt = serialLib.writeUInt

function serialLib.readUInt16(stream)
	return readUInt(stream, 16)
end

function serialLib.writeUInt16(stream, n)
	writeUInt(stream, 16, n)
end

function serialLib.readUInt32(stream)
	return readUInt(stream, 32)
end

function serialLib.writeUInt32(stream, n)
	writeUInt(stream, 32, n)
end

serialLib.readBytes = function(stream, num, bytes, start)
	bytes = bytes or {}
	start = start or 1

	for i = start, start + num - 1
		bytes[i] = readByte(stream)
	end

	return bytes
end

serialLib.writeBytes = function(stream, bytes, start, num)
	start = start or 1
	if num == nil
		num = #bytes
	end

	for i = start, start + num - 1
		writeByte(stream, bytes[i])
	end
end

serialLib.readString = function(stream)
	local len = readByte(stream)
	if len == 255
		len = serialLib.readInt32(stream)
	end

	return serialLib.bytesToString(serialLib.readBytes(stream, len))
end

serialLib.writeString = function(stream, s)
	if #s < 255
		writeByte(stream, #s)
	else
		writeByte(stream, 255)
		writeInt32(stream, #s)
	end

	serialLib.writeBytes(stream, serialLib.stringToBytes(s))
end

rawset(_G, "serialize", function(dng)
    if not dng
        return nil -- We are the client, so the return value is unimportant
    end

    local stream = serialLib.createByteStream()

	serialLib.writeByte(stream, dng.nrOfLevels) -- Write the Number of Levels
	serialLib.writeByte(stream, dng.height) -- Write Dungeon Height
	serialLib.writeByte(stream, dng.width) -- Write Dungeon Width
	--serialLib.writeByte(stream, #dng.levels) -- Dungeon level count

    for i = 1, #dng.levels
		local level = dng.levels[i]

		-- Write Dungeon Level difficulty (Enemy Spawn difficulty)
		serialLib.writeByte(stream, level.difficulty)

		-- Write whether or not the level has a Key.
		serialLib.writeByte(stream, level.hasKey)

		-- Write dimensions (Usually a perfect square)
		serialLib.writeByte(stream, #level.matrix + 1)
		
        for row = 0, #level.matrix
            for column = 0, #level.matrix[row]
                local elem = level.matrix[row][column]

                serialLib.writeByte(stream, elem.class)
                serialLib.writeByte(stream, elem.roomId)
                serialLib.writeByte(stream, elem.enemyId)
            end
        end
	end
	
	print(#stream.bytes .. " bytes of dungeon data serialized")
	
    return serialLib.bytesToString(stream.bytes)
end)

rawset(_G, "deserialize", function(serializeddng)
	print("Reading " .. #serializeddng .. " bytes of dungeon data")
	local bytes = serialLib.stringToBytes(serializeddng)
	local stream = serialLib.createByteStream(bytes)

    -- How many entries?
	local dng = {
					nrOfLevels = serialLib.readByte(stream), -- Read the Number of Levels
					height = serialLib.readByte(stream), -- Read Dungeon Height
					width = serialLib.readByte(stream), -- Read Dungeon Width
					levels = {}
				}

	--local numlevels = serialLib.readByte(stream) -- Dungeon level count

	for i = 1, dng.nrOfLevels
		dng.levels[i] = {
							height = dng.height,
							width = dng.width,
							difficulty = serialLib.readByte(stream), -- Read level difficulty
							matrix = {}
						}

		-- Read whether or not the level has a Key.
		dng.levels[i].hasKey = serialLib.readByte(stream)

		-- Read dimensions (Usually a perfect square)
		local size = serialLib.readByte(stream)

		for row = 0, size - 1
			dng.levels[i].matrix[row] = {}
			for column = 0, size - 1
				local elem = {}
				
				elem.class = serialLib.readByte(stream)
				elem.roomId = serialLib.readByte(stream)
				elem.enemyId = serialLib.readByte(stream)
				
				dng.levels[i].matrix[row][column] = elem
			end
		end
	end

	return dng
end)
