-----------------------------------------------------------------------------------
-- - - - - - - - - - - - - SRB2 Specific Functions/Hooks - - - - - - - - - - - - -- 
-----------------------------------------------------------------------------------

rawset(_G, "DNG", 0) -- Global Dungeon Table
rawset(_G, "DNGLVL", 1) -- Global Dungeon current level
rawset(_G, "SRB2matrix", {}) -- SRB2 Sector Matrix

local thing
local playeringame = 0

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

/*
This is a beta list: Doesn't Necessarily reflect how things are now.
Spindash - enemies you can beat by just spindashing in place
======
LVL 1 Crawla
LVL 1 Buzz
LVL 2 Hangster Bat (MT_HANGSTER)
LVL 3 BASH (Vulture)
LVL 4 Crawla commander

Spinjump - jump in order to defeat them
======
LVL 1 Crawla
LVL 1 Buzz
LVL 2 Pyrefly
LVL 3 Green snapper
LVL 3 Jetty-syn bomber
LVL 4 Jetty-syn gunner

Flank/Dodge - deliberate avoidance of their attack is often necessary
======
LVL 1 Pointy (MT_POINTY)
LVL 2 Egg guard
LVL 2 Minus
LVL 3 Pop-up turret
LVL 3 Robo-hood
LVL 3 Spincushion
LVL 4 Lance-a-bot
LVL 4 Jetty-syn gunner
LVL 5 Deton
LVL 5 Hive Elemental
*/

rawset(_G, "enemyTable", {
	-- Spindash
	[1] = {
		-- Difficulty Levels
		[1] = {
				MT_BLUECRAWLA,
				MT_GOLDBUZZ,
				},
		[2] = {
				MT_HANGSTER
				},
		[3] = {
				MT_VULTURE
				},
		[4] = {
				MT_NULL -- Freebie!
				},
		--[5] = { }
	},
	
	--Spin Jump
	[2] = {
		-- Difficulty Levels
		[1] = {
				MT_REDCRAWLA,
				MT_REDBUZZ
				},
		[2] = {
				MT_PYREFLY
				},
		[3] = {
				MT_GSNAPPER,
				MT_JETTBOMBER
				},
		[4] = {
				MT_JETTGUNNER,
				MT_POPUPTURRET,
				MT_CRAWLACOMMANDER
				},
		--[5] = { },
	},
	
	-- Flank/Dodge
	[3] = {
		-- Difficulty Levels
		[1] = {
				MT_NULL -- Freebie!
				},
		[2] = {
				MT_MINUS,
				},
		[3] = {
				MT_CACOLANTERN,
				MT_ROBOHOOD
				},
		[4] = {
				MT_FACESTABBER,
				MT_BUMBLEBORE,
				--MT_DRAGONBOMBER,
				MT_JETTGUNNER
				},
		[5] = {
				MT_DETON,
				MT_HIVEELEMENTAL
				},
	}
})

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

rawset(_G, "toSRB2matrix", function(level)
	-- Let's begin!
	-- Put the first 2303 sectors in a matrix

	local j, k = 0, 0
	SRB2matrix[j] = {}
	for i = 0, (2304-1) do -- Dumb Matrix
		k = (i%(level.width+1))
		SRB2matrix[j][k] = sectors[i]
		--print("sector["..i.."] indexed at ("..j..","..k..")")

		if (k==level.width)
			j = $ + 1
			SRB2matrix[j] = {}
		end
	end
end)

rawset(_G, "goToLevel", function(level)
	if DNG
		local gridsize = 256<<FRACBITS	-- grid size in map units
		local tblrt = P_RandomRange(1, #enemyTable) -- Table room type (SEE: Tile.ENEMY below). Make it consistent.
		
		-- Great, we have a matrix to compare with! Let's start assigning sector blocks!
		for j=0,level.width do
			for i=0,level.height do
				local tile = level.matrix[i][j]
				local sec = SRB2matrix[i][j]
				local sx, sy = j*gridsize+(128<<FRACBITS), -i*gridsize-(128<<FRACBITS) -- Spawn coords
				
				if (tile.class == Tile.WALL)
					sec.ceilingheight = sec.floorheight
					sec.special = 0 -- Nothing special with this sector
				elseif (tile.class == Tile.WALK) or (tile.class == Tile.CORR)
					sec.special = 0 -- Nothing special with this sector
				elseif (tile.class == Tile.SPAWN)
					sec.floorpic = "SPACEYL"
					sec.floorheight = 16<<FRACBITS	
					sec.special= 0 -- Nothing special with this sector

					--local sx, sy = 0,0 -- Spawn coords (Old method)
					for l=0, #sec.lines-1 do
						--sx = $ + (sec.lines[l].v1.x + sec.lines[l].v2.x)/2 -- Midpoint X
						--sy = $ + (sec.lines[l].v1.y + sec.lines[l].v2.y)/2 -- Midpoint Y
						sec.lines[l].frontside.bottomtexture = R_TextureNumForName(sec.floorpic)
						sec.lines[l].backside.bottomtexture = R_TextureNumForName(sec.floorpic)
					end
					--sx = $/#sec.lines -- Center X
					--sy = $/#sec.lines -- Center Y
					
					-- This is a failsafe. In the event that DNG is deleted haphazardly for whatever reason.
					-- Also spectators I guess???
					for mapt in mapthings.iterate
						if (mapt.type <= 32) -- Player spawns
							mapt.x = sx / FRACUNIT
							mapt.y = sy / FRACUNIT
							mapt.z = sec.floorheight / FRACUNIT
						end
					end
					for p in players.iterate
						if p.mo and p.mo.valid and not p.spectator
							P_TeleportMove(p.mo, sx, sy, sec.floorheight)
						end
					end	
				elseif (tile.class == Tile.SPRING)
					sec.floorpic = "SPACEGL"
					sec.floorheight = 16<<FRACBITS
					sec.special = 0 --2<<12 -- Exit Sector/Special Stage Pit/Return Flag
					
					--local sx, sy = 0,0 -- Spawn coords (Old method)
					for l=0, #sec.lines-1 do
						--sx = $ + (sec.lines[l].v1.x + sec.lines[l].v2.x)/2 -- Midpoint X
						--sy = $ + (sec.lines[l].v1.y + sec.lines[l].v2.y)/2 -- Midpoint Y
						sec.lines[l].frontside.bottomtexture = R_TextureNumForName(sec.floorpic)
						sec.lines[l].backside.bottomtexture = R_TextureNumForName(sec.floorpic)
					end
					--sx = $/#sec.lines -- Center X
					--sy = $/#sec.lines -- Center Y
					
					local spring = P_SpawnMobj(sx, sy, sec.floorheight, MT_REDSPRING)
					--sign.angle = Level:getCenter()
				elseif (tile.class == Tile.RING)
					local dist = 64*FRACUNIT
					sec.floorheight = 0
					--sec.floorpic = "SPACEYL"
					P_SpawnMobj(sx, sy, sec.floorheight+16*FRACUNIT, MT_RING) -- Middle
					P_SpawnMobj(sx-dist, sy, sec.floorheight+16*FRACUNIT, MT_RING)
					P_SpawnMobj(sx+dist, sy, sec.floorheight+16*FRACUNIT, MT_RING)
					P_SpawnMobj(sx, sy-dist, sec.floorheight+16*FRACUNIT, MT_RING)
					P_SpawnMobj(sx, sy+dist, sec.floorheight+16*FRACUNIT, MT_RING)
				elseif (tile.class == Tile.ITEM)
					sec.floorheight = 0
					-- At lower difficulties, a higher chance for a shield to spawn.
					-- At higher difficulties, a lower chance for a shield to spawn.
					local itm = P_RandomChance(FRACUNIT/(level.difficulty*2)) and P_RandomRange(MT_RING_BOX, MT_ELEMENTAL_BOX) or MT_RING_BOX
					
					P_SpawnMobj(sx, sy, sec.floorheight, itm)
				elseif (tile.class == Tile.ENEMY)
					sec.floorheight = 0
					--sec.floorpic = "SPACERL"
					local enemy -- Enemy that will spawn
					--local tblrc -- Table room choice. Defaults to 1. (Defined above)
					local tbldc = 1 -- Table difficulty choice. Defaults to 1.
					local tblec = 1 -- Table enemy choice. Defaults to 1.

					if tile.roomId -- This is a room
						if tile.enemyId -- Room has a pecific enemyId?
							tblrt = tile.enemyId -- Room Type = room enemyId
						end
						
						if (level.difficulty > 1)
							if (enemyTable[tblrt][level.difficulty] ~= nil) -- Difficulty exists for given table value
								tbldc = P_RandomRange(level.difficulty-1, level.difficulty)
							else -- Difficulty doesn't exist for given table value
								tbldc = P_RandomRange(level.difficulty-1, #enemyTable[tblrt])
							end
						end
						tblec = P_RandomRange(1, #enemyTable[tblrt][tbldc])
					else -- No room ID associated with this tile. Likely a corridor.
						-- Make simple enemies
						tbldc = P_RandomRange(1,2)
						tblec = P_RandomRange(1, #enemyTable[tblrt][tbldc])
					end
					
					enemy = P_SpawnMobj(sx, sy, sec.floorheight,
											enemyTable[tblrt][tbldc][tblec]
											)
					enemy.angle = P_RandomRange(0,35)*ANG10 + ANGLE_180
					if (enemy.flags & MF_NOGRAVITY)
						enemy.z = $ + 196<<FRACBITS
					end
					if (enemy.flags & MF_SPAWNCEILING)
						enemy.z = sec.ceilingheight - enemy.height
					end
				elseif (tile.class == Tile.KEY)
					sec.floorheight = 0
					--sec.floorpic = "SPACEYL"
					
					local key = P_SpawnMobj(sx, sy, 16<<FRACBITS, MT_DNGKEY)
				elseif (tile.class == Tile.BOSS)
					sec.floorpic = "SPACERL"
					sec.floorheight = 16<<FRACBITS
					sec.special = 0 -- Nothing special with this sector
					
					--local sx, sy = 0,0 -- Spawn coords (Old method)
					for l=0, #sec.lines-1 do
						--sx = $ + (sec.lines[l].v1.x + sec.lines[l].v2.x)/2 -- Midpoint X
						--sy = $ + (sec.lines[l].v1.y + sec.lines[l].v2.y)/2 -- Midpoint Y
						
						sec.lines[l].frontside.bottomtexture = R_TextureNumForName(sec.floorpic)
						sec.lines[l].backside.bottomtexture = R_TextureNumForName(sec.floorpic)
					end
					--sx = $/#sec.lines -- Center X
					--sy = $/#sec.lines -- Center Y
					
					local boss = P_SpawnMobj(sx, sy, 64*FRACUNIT, MT_EGGMOBILE)
				elseif (tile.class == Tile.PIT)
					sec.floorpic = "PIT"
					sec.floorheight = -(256<<FRACBITS)
					
					for l=0, #sec.lines-1 do
						sec.lines[l].frontside.bottomtexture = R_TextureNumForName("SPACERL")
						sec.lines[l].backside.bottomtexture = R_TextureNumForName("SPACERL")
					end

					sec.special = SD_DEATHPITNOTILT
					sec.damagetype = SD_DEATHPITNOTILT -- Death Pit (No Camera Tilt)
				end
			end
		end
	else
		error("No specified dungeon exists! Did you call goToLevel before toSRB2matrix?")
	end
end)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

addHook("MapChange", function(mapnum)
	if not mapheaderinfo[mapnum].dungeon -- Consider this in line with 'leaving a dungeon'
		DNGLVL = 1
	end
end)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

local gamestateloaded = false
addHook("PreThinkFrame", do
	if mapheaderinfo[gamemap].dungeon and gamestateloaded
		DNG = deserialize(DNG)
		gamestateloaded = false
	end
	
	-- Prevent player joins if dungeon level matrix is too large
	local status, allowjoin = pcall(CV_FindVar, "allowjoin")
	if status and mapheaderinfo[gamemap].dungeon and DNG and #DNG.levels > 100 and allowjoin
		COM_BufInsertText(server, "allowjoin off")
	end

	playeringame = 0
	for p in players.iterate
		if p.spectator continue end -- We're a spectator. Skip.
		if not p.mo continue end -- Not a mo object. Skip.
		if p.bot continue end  -- Player is a bot. Skip.
		playeringame = $ + 1
	end
end)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

addHook("PlayerSpawn", function(p)
	if p and p.valid
	and mapheaderinfo[gamemap].dungeon -- Is this a dungeon?
		p.haskey = 0
		p.keystatus = 0
		
		-- Give the player what they had on the previous floor!		
		if p.savedrings
			p.rings = p.savedrings
			-- p.xtralife can only be a max of 2.
			-- Fix players abruptly gaining extra lives each floor load
			p.xtralife = p.rings/100
			p.savedrings = nil
		end
		if p.savedscore
			p.score = p.savedscore
			p.savedscore = nil
		end
		if p.savedshield
			p.powers[pw_shield] = p.savedshield
			P_SpawnShieldOrb(p)
			p.savedshield = nil
		end
		
		-- Late joiners.
		if p.mo and p.mo.valid
		and DNG	-- Has the dungeon already been initialized from the gamestate?
		and (leveltime > 0)
			local mo = p.mo
			local level = DNG.levels[DNGLVL]
			local gridsize = 256<<FRACBITS	-- grid size in map units

			--toSRB2matrix(level)

			for i=0,#level.matrix do
				for j=0,#level.matrix[i] do
					local tile = level.matrix[i][j]
					--local sec = SRB2matrix[i][j]
					local sx, sy = j*gridsize+(128<<FRACBITS), -i*gridsize-(128<<FRACBITS) -- Spawn coords
					
					if (tile.class == Tile.SPAWN)					
						-- This is a failsafe. In the event that DNG is deleted haphazardly for whatever reason.
						-- Also spectators I guess???
						for mapt in mapthings.iterate
							if (mapt.type <= 32) -- Player spawns
								mapt.x = sx / FRACUNIT
								mapt.y = sy / FRACUNIT
								mapt.z = 0--sec.floorheight / FRACUNIT
							end
						end
					
						P_TeleportMove(p.mo, sx,sy,0)
						break -- We're done here. Go home!
					end
				end
			end
		end
	end
end)


addHook("MobjDeath", function(mo)
	if mo and mo.valid
	and mo.player and mo.player.valid
		local p = mo.player -- Simplify
		
		-- Player has Key upon death?
		if p.haskey
		and DNG	-- Has the dungeon already been initialized from the gamestate?
		and DNG.levels
		and DNG.levels[DNGLVL].hasKey
			for px in players.iterate
				CONS_Printf(px, "Player \x82"..p.name.."\x80 died while holding the key. Respawning key...")
			end
			p.haskey = false
			
			local level = DNG.levels[DNGLVL]
			local gridsize = 256<<FRACBITS	-- grid size in map units
			
			for i=0,#level.matrix do
				for j=0,#level.matrix[i] do
					local tile = level.matrix[i][j]
					local sx, sy = j*gridsize+(128<<FRACBITS), -i*gridsize-(128<<FRACBITS) -- Spawn coords
					
					if (tile.class == Tile.KEY)
						local key = P_SpawnMobj(sx, sy, 16<<FRACBITS, MT_DNGKEY)
						P_SpawnMobj(sx, sy, key.z, MT_EXPLODE) -- Dust cloud
						S_StartSound(key, sfx_s3k9c) -- [Got Super Emerald, SFX]
						key.scale = 1
						key.destscale = FRACUNIT
						key.scalespeed = FRACUNIT/TICRATE
						break
					end
				end
			end
		end
	end
end, MT_PLAYER)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

-- Dungeon Key Stuff
freeslot("MT_DNGKEY", "S_DNGKEY", "SPR_DKEY")

thing = MT_DNGKEY
mobjinfo[thing].doomednum = 3025
mobjinfo[thing].spawnstate = S_DNGKEY
mobjinfo[thing].spawnhealth = 1
mobjinfo[thing].deathstate = S_DNGKEY
mobjinfo[thing].raisestate = S_DNGKEY
mobjinfo[thing].radius = 24*FRACUNIT
mobjinfo[thing].height = 48*FRACUNIT
mobjinfo[thing].flags = MF_SPECIAL|MF_NOGRAVITY

thing = S_DNGKEY
states[thing].sprite = SPR_DKEY
states[thing].frame = A|FF_PAPERSPRITE
states[thing].tics = -1
states[thing].nextstate = S_DNGKEY

addHook("MobjSpawn", function(mo)
	mo.zangle = 0
end, MT_DNGKEY)

-- What happens when you touch the key?
addHook("TouchSpecial", function(mobj, toucher)
	local sound = sfx_None
	local temp
	
	if not toucher.player and not toucher.player.valid
		return
	end

	local player = toucher.player
	
	if player.spectator
		return
	end
	
	if not P_CanPickupItem(player, true)
		return
	end

	sound = sfx_kc5e
	
	temp = P_SpawnMobj(mobj.x, mobj.y, mobj.z, mobj.type)
	temp.flags = $ | (MF_NOGRAVITY) & ~MF_SPECIAL
	temp.health = 0
	temp.fuse = 2*TICRATE --2*TICRATE/3 -- Die in 2/3's of a second
	temp.angle = mobj.angle
	P_SetMobjStateNF(temp, mobjinfo[mobj.type].raisestate)
	temp.frame = $ | TR_TRANS50
	temp.destscale = 1
	temp.scalespeed = FRACUNIT/60
	
	player.haskey = true

	P_SetMobjStateNF(mobj, S_NULL)
	if (sound)
		S_StartSound(player.mo, sound)
	end
end, MT_DNGKEY)

-- Bobbing and Dissapearing Thinker
addHook("MobjThinker", function(mo)
	if mo and mo.valid
	and mo.health
		mo.zangle = $ + ANG10
		mo.z = $ + FixedMul(sin(mo.zangle), 2*FRACUNIT) -- Bob
		mo.angle = $ + ANG15/2
		if not (leveltime % 5)
			local ang = ANGLE_45*P_RandomRange(1,8)
			local sparkle = P_SpawnMobj(mo.x, mo.y, 16<<FRACBITS, MT_NIGHTSPARKLE)
			if mo.eflags & MFE_VERTICALFLIP
				sparkle.momz = -2*FRACUNIT
			else
				sparkle.momz = 2*FRACUNIT
			end
			sparkle.angle = ang
			P_InstaThrust(sparkle, sparkle.angle, FRACUNIT*2)
			P_SetMobjStateNF(sparkle, S_NIGHTSPARKLE2)
		end
	elseif mo and mo.valid -- Fading key
		mo.angle = $ + ANG30
		P_SetObjectMomZ(mo, FRACUNIT, false)
		
		if not (leveltime % 5)
			local r = mobjinfo[mo.type].radius>>FRACBITS
			local sparkle = P_SpawnMobj(mo.x + (P_RandomRange(r, -r)<<FRACBITS),
										mo.y + (P_RandomRange(r, -r)<<FRACBITS),
										mo.z + (P_RandomKey(mobjinfo[mo.type].height>>FRACBITS)<<FRACBITS),
										MT_NIGHTSPARKLE)
			if mo.eflags & MFE_VERTICALFLIP
				sparkle.momz = -2*FRACUNIT
			else
				sparkle.momz = 2*FRACUNIT
			end
			P_SetMobjStateNF(sparkle, S_NIGHTSPARKLE2)
		end
	end
end, MT_DNGKEY)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

-- Aura for the Spring being locked
freeslot("MT_LOCKAURA", "S_LOCKAURA", "SPR_NTHK")
freeslot("sfx_lockbr", "sfx_lockfa")

thing = MT_LOCKAURA
mobjinfo[thing].doomednum = -1
mobjinfo[thing].spawnstate = S_LOCKAURA
mobjinfo[thing].radius = 48*FRACUNIT
mobjinfo[thing].height = 48*FRACUNIT
mobjinfo[thing].flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIPHEIGHT

thing = S_LOCKAURA
states[thing].sprite = SPR_NTHK
states[thing].frame = A|FF_TRANS30
states[thing].tics = -1
states[thing].nextstate = S_LOCKAURA

thing = sfx_lockbr
sfxinfo[thing].caption = "Spring Unlock"
sfxinfo[thing].flags = SF_X4AWAYSOUND|SF_X8AWAYSOUND

thing = sfx_lockfa
sfxinfo[thing].caption = "Spring Unlock burst"
sfxinfo[thing].flags = SF_X4AWAYSOUND|SF_X8AWAYSOUND

addHook("MobjThinker", function(mo)
	if mo and mo.valid
		mo.threshold = $ or 1
		mo.pulsefade = $ or false
		
		mo.frame = A|(FF_TRANS10 * mo.threshold)

		if not (leveltime%2)
			if not mo.pulsefade
				mo.threshold = $ + 1
			else
				mo.threshold = $ - 1
			end
			
			if (mo.threshold > 7)
				mo.pulsefade = true
			elseif (mo.threshold < 3)
				mo.pulsefade = false
			end
		end

		if not (leveltime%5)
			local r = mo.radius>>FRACBITS
			local sparkle = P_SpawnMobj(mo.x + (P_RandomRange(r, -r)<<FRACBITS),
										mo.y + (P_RandomRange(r, -r)<<FRACBITS),
										mo.z + mo.height/2,
										MT_NIGHTSPARKLE)
			if mo.eflags & MFE_VERTICALFLIP
				sparkle.momz = -2*FRACUNIT
			else
				sparkle.momz = 2*FRACUNIT
			end
			P_SetMobjStateNF(sparkle, S_NIGHTSPARKLE2)
		end
	end
end, MT_LOCKAURA)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

-- Sign stuff
addHook("MobjThinker", function(mo)
	if mo and mo.valid
	and mapheaderinfo[gamemap].dungeon -- Is this a dungeon?
		if P_IsObjectOnGround(mo)
			mo.angle = R_PointToAngle(mo.x, mo.y) -- Always face the camera
			--for p in players.iterate
			--	if p.spectator continue end -- We're a spectator. Skip.
			--	if (p.pflags & PF_FINISHED) continue end -- We've finished the level. Skip.
			--end
		else
			mo.angle = $ - ANG2
		end
	end
end, MT_SIGN)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

-- Eggmobile death stuff
addHook("MobjDeath", function(mo)
	-- Eggman died! Let's end this Dungeon Crawler!
	if mo and mo.valid
	and mapheaderinfo[gamemap].dungeon -- Is this a dungeon?
	and DNG	-- Has the dungeon already been initialized?
		local level = DNG.levels[DNGLVL]
		local gridsize = 256<<FRACBITS	-- grid size in map units
		for i=0,level.height do
			for j=0,level.width do
				local tile = level.matrix[i][j]
				local sec = SRB2matrix[i][j]
				local sx, sy = j*gridsize+(128<<FRACBITS), -i*gridsize-(128<<FRACBITS) -- Spawn coords
				
				if (tile.class == Tile.BOSS)
					sec.floorpic = "SPACEGL"
					sec.floorheight = 16<<FRACBITS
					
					for l=0, #sec.lines-1 do
						sec.lines[l].frontside.bottomtexture = R_TextureNumForName(sec.floorpic)
						sec.lines[l].backside.bottomtexture = R_TextureNumForName(sec.floorpic)
					end
					
					local sign = P_SpawnMobj(sx, sy, sec.ceilingheight - mo.height, MT_SIGN)
					S_StartSound(sign, sfx_lvpass)
					sec.special = 2<<12 -- Exit Sector/Special Stage Pit/Return Flag
					break -- We're done here. Go home!
				end
			end
		end
	end
end, MT_EGGMOBILE)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

-- Spring stuff
addHook("MobjSpawn", function(mo)
	if mapheaderinfo[gamemap].dungeon -- Is this a dungeon?
		mo.savedflags = mo.flags
		mo.locked = false
		mo.flags = MF_NOCLIP
		mo.radius = $*2
		mo.scale = $*3
	end
end, MT_REDSPRING)

--
-- getMobjForType(type)
--
-- Description:
-- Return the first mobj that is found with the right type.
-- Useful for finding a single unique object in a map.
-- If no map object was found, it returns nil.
--
local function getMobjForType(type)
	for i_mo in mobjs.iterate() do -- i_mo is "iterating mo"
		if (i_mo.type != type) -- Not the droid we're looking for.
			continue -- Skip it.
		end
		
		if (i_mo.health <= 0) -- Dead?
			continue -- Skip it.
		end
		
		return i_mo
	end
	return nil
end

addHook("MobjThinker", function(spring)
	if spring and spring.valid
	and spring.savedflags
	and mapheaderinfo[gamemap].dungeon -- Is this a dungeon?
		if spring.locked -- Level Clear condition
			spring.frame = A|FF_TRANS50
			
			if not spring.aura
				spring.aura = P_SpawnMobj(spring.x, spring.y, spring.z - (mobjinfo[MT_LOCKAURA].height), MT_LOCKAURA)
				spring.aura.color = SKINCOLOR_WHITE
				spring.aura.target = spring
				spring.aura.scale = spring.scale
			end
			
			for p in players.iterate do
				if p.spectator continue end -- We're a spectator. Skip.
				if not p.mo continue end -- Not a mo object. Skip.
				if (P_AproxDistance(P_AproxDistance(p.mo.x - spring.x, p.mo.y - spring.y), 
													p.mo.z - spring.z) > 4*spring.radius/5)
					continue -- Out of range
				end
				
				-- Unlocking Anim
				if p.haskey
					-- Spawn Sparks
					for i = 1, 8
						local fa = i*ANGLE_45
						local spark = P_SpawnMobj(spring.x, spring.y, spring.z + spring.height/2, MT_SUPERSPARK)
						spark.angle = fa
						spark.fuse = 10 -- 5 frames, 2 tics each
						P_InstaThrust(spark, spark.angle, 36*FRACUNIT)
						P_FlashPal(p, 2, TICRATE/6)
					end
					S_StopSound(spring) -- Stop whatever sound the spring may be playing.
					
					S_StartSound(spring, sfx_lockbr) -- Unlock the spring!
					if not S_SoundPlaying(spring, sfx_lockfa) -- Avoid loud and unnecesary sound.
						S_StartSound(spring, sfx_lockfa)
					end
					p.haskey = false
					
					-- Nested Iterator... I know...
					-- Whomever the player was that had the key (p.haskey) would be the only player to activate the code under p.haskey...
					-- I wanted to apply the current key status to all players,
					-- and this was the easiest method to do it even if it doesn't seem clean.
					for px in players.iterate
						CONS_Printf(px, "\x82" ..p.name.."\x80 has unlocked the spring!")
						px.keystatus = 1 -- Unlocked
					end
					spring.aura.fuse = 1
					spring.aura = nil
					spring.locked = false
				else -- No key?
					if not (leveltime%(TICRATE*2)) -- Every few seconds...
						S_StartSoundAtVolume(spring, sfx_s3kb2, 175) -- Play an annoying sound!
					end
				end
			end
		else
			spring.frame = A
			
			-- Code is called in the thinker because a key will not always be found on spawn.
			spring.locked = getMobjForType(MT_DNGKEY) and true or false
			
			spring.ptouch = 0
			for p in players.iterate do
				if p.spectator continue end -- We're a spectator. Skip.
				if not p.mo continue end -- Not a mo object. Skip.
				if (P_AproxDistance(P_AproxDistance(p.mo.x - spring.x, p.mo.y - spring.y), 
													p.mo.z - spring.z) > 4*spring.radius/5)
					continue -- Out of range
				end
				spring.ptouch = $ + 1
			end
			
			-- This is a hack because MobjCollide doesn't want to work with me.
			if spring.ptouch
			and (spring.ptouch == playeringame)
				spring.flags = spring.savedflags
				spring.savedflags = 0
				
				--print("Move on to the next level!")
				for p in players.iterate
					if not p.mo continue end -- Not a mo object
					if p.bot continue end -- Player is a bot. Skip.
					local mo = p.mo -- Simplify ourselves.
					local statch, camdist = pcall(CV_FindVar, "cam_dist")
					local statcd, camheight = pcall(CV_FindVar, "cam_height")
					if statch and statcd and camdist and camheight
						camdist = camdist.value/FRACUNIT
						camheight = camheight.value/FRACUNIT
					else
						camdist = 250
						camheight = 50
					end
					local cam = P_SpawnMobj(spring.x + FixedMul(cos(mo.angle), -(3*camdist*FRACUNIT)/2),
											spring.y + FixedMul(sin(mo.angle), -(3*camdist*FRACUNIT)/2),
											spring.z + P_GetPlayerHeight(p) + camheight*FRACUNIT,
											MT_THOK)
					cam.angle = R_PointToAngle2(cam.x, cam.y, spring.x, spring.y)
					cam.flags = $ | MF2_DONTDRAW
					cam.tics = -1
					cam.fuse = -1
					p.awayviewmobj = cam
					p.awayviewtics = -1

					mo.momx = 0
					mo.momy = 0
					p.savedrings = p.rings -- Store the current rings somewhere nice!
					p.savedscore = p.score
					p.savedshield = p.powers[pw_shield]
					mo.flags = $ | MF_NOGRAVITY
					if DNG and #DNG.levels and (DNGLVL < #DNG.levels) -- Cheeky check for those that try to hack the system. :)
						G_SetCustomExitVars(gamemap, 1) -- Nextmap will be the same map
					end
					P_DoPlayerExit(p)
				end
				if DNG -- Don't even increment if it doesn't exist...
					if #DNG.levels and (DNGLVL < #DNG.levels)
						DNGLVL = $ + 1
					else
						DNGLVL = 1
					end
				end
			end
		end
		return true
	end
end, MT_REDSPRING)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

COM_AddCommand("dng_printlevel", function(p, lvl)
	if not lvl then	CONS_Printf(p, "printlevel <num>: prints the output of a given dungeon level (num)") return end
	if not mapheaderinfo[gamemap].dungeon then CONS_Printf(p, "Sorry. This command can only be used in dungeon maps!") return end
	if not DNG then error("Dungeon Doesn't exist!") end
	-- DNG is valid from this point forward
	if not tonumber(lvl) then error("Desired input is invalid") end
	lvl = tonumber($)
	if (lvl > #DNG.levels) or (lvl < 1) then error("Desired level cannot be printed: Doesn't exist.") end

	local s = "L E V E L  "..lvl
	print(s) -- Eg. Level 1 prints "L E V E L  1"
	
	lvl = DNG.levels[lvl] -- Manipulate this number into a dungeon level object
	for i=0,lvl.height do
		local row=""
		for j=0,lvl.width do
			row=row..lvl.matrix[i][j].class..","
			--row=row.."("..i..","..j..")"..","
			--row=row..lvl.matrix[i][j].roomId..","    -- for exposing room-ids
			--row=row..lvl.matrix[i][j].enemyId..","    -- for exposing enemytype-ids
		end
		print(row)
	end
end, COM_LOCAL)


rawset(_G, "cv_dnglvl", CV_RegisterVar({
	name = "dng_levels",
	defaultvalue = 10,
	flags = CV_NETVAR|CV_CALL,
	PossibleValue = {MIN = 1, MAX = 999},
	Func = function()
		print("\x82".."NOTICE: ".."\x80".."Dungeon Levels have been modified!") -- set to "..cv_dnglvl.value)
		print("Please exit and reload the current dungeon (Ex. map mapXX) for changes to take effect.")
	end,
}))

COM_AddCommand("dng_clear", function(p)
	CONS_Printf(p, 'Clearing the current dungeon...')
	DNG = 0
	DNGLVL = 1
	SRB2matrix = {}
	if mapheaderinfo[gamemap].dungeon
		CONS_Printf(p, "Please reload the current dungeon (Ex. map mapXX) for changes to take effect.")
	end
end, 1) -- Server Admin only

COM_AddCommand("dng_skiplevel", function(p)
	if not mapheaderinfo[gamemap].dungeon then CONS_Printf(p, "Sorry. This command can only be used in dungeon maps!") return end
	if DNG and (DNGLVL == #DNG.levels) then CONS_Printf(p, "Sorry! Cannot skip the last dungeon level!") return end
	
	for i_p in players.iterate
		local mo = i_p.mo
		for thing in mobjs.iterate()
			if (thing.type != MT_REDSPRING) continue end -- Not a Red Spring? Skip it.
			if (thing.health <= 0) continue end -- Dead for any reason
			
			-- Teleport to it.
			P_TeleportMove(mo, thing.x, thing.y, thing.z + thing.height/2)
			CONS_Printf(i_p, "\x82"..p.name.."\x80 skipped the level!")
		end
	end
end, 1) -- Server Admin only

COM_AddCommand("dng_gotolevel", function(p, lvl)
	if not mapheaderinfo[gamemap].dungeon then CONS_Printf(p, "Sorry. This command can only be used in dungeon maps!") return end
	if not DNG then error("Dungeon Doesn't exist!") end
	if not tonumber(lvl) then error("Desired input is invalid!") end
	lvl = tonumber($)
	if (lvl > #DNG.levels) or (lvl < 1) then error("Desired level cannot be skipped to: Doesn't exist.") end
	DNGLVL = lvl
	
	CONS_Printf(p, "Warping to Dungeon Level...")
	G_SetCustomExitVars(gamemap, 1) -- Nextmap will be the same map
	G_ExitLevel()
end, 1)

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- 

hud.add(function(v, p)
	if v
	and p and p.valid and not p.spectator
	and mapheaderinfo[gamemap].dungeon
	--and DNG
		local x, y = 16, 12*5 - 2 --(v.width()/v.dupx()-30), 12 --(v.height()/v.dupy())
		local dnum
		local dlvl = v.cachePatch("STTLEVEL")
		local dslsh = v.cachePatch("STTSLASH")
		local dlock = { v.cachePatch("STTLOCK1"), v.cachePatch("STTLOCK2") }
		local dkey = v.cachePatch("STTKEY1")
		
		local vidflags = V_SNAPTOTOP|V_SNAPTOLEFT
		
		-- Graphical fix. Don't mistakenly display the wrong number if exiting
		if (p.exiting)
		and DNG and #DNG.levels 
		and (DNGLVL ~= #DNG.levels)
			dnum = DNGLVL-1
		else
			dnum = DNGLVL
		end
		
		-- "LEVEL" Text
		v.draw(x,y,dlvl,vidflags)
		
		-- Current DNGLVL
		--Adjust the x offset
		x = ($ * 6) - 8
		v.drawNum(x-2,y,dnum,vidflags)
		v.draw(x,y,dslsh,vidflags)
		
		-- Max dungeon levels
		x = $ + 32
		if DNG and #DNG.levels
			v.drawNum(x,y,#DNG.levels,vidflags)
		else
			v.drawString(x,y+2,"???",vidflags,"right")
		end
		
		-- Lock and Key
		if DNG and DNG.levels[DNGLVL].hasKey and not p.exiting
			x, y = $ + 4, $ - 2
			if p.haskey
				if not (leveltime%2)
					v.draw(x,y,dlock[p.keystatus+1],vidflags) -- Lock
				end
				
				x = $ + 18
				v.draw(x,y,dkey,vidflags) -- Key
			else
				v.draw(x,y,dlock[p.keystatus+1],vidflags) -- Lock
			end
		end
	end
end, "game")

-- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### -- ##### --

addHook("NetVars", function(n)
	if mapheaderinfo[gamemap].dungeon
		DNGLVL = n(DNGLVL)
		
		local serialised = serialize(DNG)
		serialised = n($) -- Grab serialized DNG from gamestate.
		DNG = serialised
		gamestateloaded = true -- Go to PreThinkFrame now.
	end
end)